/*********************************************************************
 *
 * Software License Agreement (BSD License)
 *
 *  Copyright (c) 2008, Willow Garage, Inc.
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of the Willow Garage nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 *
 *  \author Sachin Chitta, Ioan Sucan
 *********************************************************************/

#include "planning_environment/monitors/environment_server_setup.h"
#include "planning_environment/monitors/planning_monitor.h"

#include <trajectory_msgs/JointTrajectory.h>
#include <sensor_msgs/JointState.h>
#include <motion_planning_msgs/DisplayTrajectory.h>

#include <planning_environment_msgs/GetRobotState.h>
#include <planning_environment_msgs/GetJointTrajectoryValidity.h>
#include <planning_environment_msgs/GetStateValidity.h>
#include <planning_environment_msgs/GetJointsInGroup.h>
#include <planning_environment_msgs/GetEnvironmentSafety.h>
#include <planning_environment_msgs/SetConstraints.h>
#include "planning_environment/monitors/environment_server_setup.h"
#include <valarray>
#include <visualization_msgs/Marker.h>

#include <visualization_msgs/MarkerArray.h>

namespace planning_environment
{
class EnvironmentServer
{
public:	
  EnvironmentServer(EnvironmentServerSetup &setup) : private_handle_("~"),setup_(setup)
  {
    int velocity_history_size;
    private_handle_.param<int>("velocity_history_size", velocity_history_size, 20);

    joint_state_subscriber_ =  root_handle_.subscribe("joint_states", 1, &planning_environment::EnvironmentServer::jointStateCallback, this);

    vis_marker_publisher_ = root_handle_.advertise<visualization_msgs::Marker>("environment_server_contact_markers", 128);
 	    
    planning_monitor_ = setup_.planning_monitor_;
    planning_monitor_->getEnvironmentModel()->setVerbose(false);

    tf_              = &setup_.tf_;

    get_trajectory_validity_service_ = private_handle_.advertiseService("get_trajectory_validity", &EnvironmentServer::getTrajectoryValidity, this);
    get_env_safety_service_ = private_handle_.advertiseService("get_environment_safety", &EnvironmentServer::getEnvironmentSafety, this);
    get_execution_safety_service_ = private_handle_.advertiseService("get_execution_safety", &EnvironmentServer::getTrajectoryValidity, this);
    set_constraints_service_ = private_handle_.advertiseService("set_constraints", &EnvironmentServer::setConstraints, this);
    get_joints_in_group_service_ = private_handle_.advertiseService("get_joints_in_group", &EnvironmentServer::getJointsInGroup, this);

    get_robot_state_service_ = private_handle_.advertiseService("get_robot_state", &EnvironmentServer::getRobotState, this);
    get_state_validity_service_ = private_handle_.advertiseService("get_state_validity", &EnvironmentServer::getStateValidity, this);
    allowed_contact_regions_publisher_ = private_handle_.advertise<visualization_msgs::MarkerArray>("allowed_contact_regions_array", 128);

    velocity_history_.resize(velocity_history_size);
    velocity_history_index_ = 0;

    ROS_INFO("Environment server started");
  }
	
  virtual ~EnvironmentServer()
  {
  }
	
private:
		
  /// Representation of a state and its corresponding joint constraints
  struct StateAndConstraint
  {
    boost::shared_ptr<planning_models::KinematicState> state;
    std::vector<motion_planning_msgs::JointConstraint> constraints;	    
  };
	
	
  /** @brief The ccost and display arguments should be bound by the caller. 
   * This is a callback function that gets called by the planning
   * environment when a collision is found */
  void contactFound(collision_space::EnvironmentModel::Contact &contact)
  {
    static int count = 0;

    std::string ns_name;
        
    if(contact.link1 != NULL) {
      //ROS_INFO_STREAM("Link 1 is " << contact.link2->name);
      if(contact.link1_attached_body_index == 0) {
        ns_name += contact.link1->name+"+";
      } else {
        if(contact.link1->attachedBodies.size() < contact.link1_attached_body_index) {
          ROS_ERROR("Link doesn't have attached body with indicated index");
        } else {
          ns_name += contact.link1->attachedBodies[contact.link1_attached_body_index-1]->id+"+";
        }
      }
    } 
    
    if(contact.link2 != NULL) {
      //ROS_INFO_STREAM("Link 2 is " << contact.link2->name);
      if(contact.link2_attached_body_index == 0) {
        ns_name += contact.link2->name;
      } else {
        if(contact.link2->attachedBodies.size() < contact.link2_attached_body_index) {
          ROS_ERROR("Link doesn't have attached body with indicated index");
        } else {
          ns_name += contact.link2->attachedBodies[contact.link2_attached_body_index-1]->id;
        }
      }
    } 
    
    if(!contact.object_name.empty()) {
      //ROS_INFO_STREAM("Object is " << contact.object_name);
      ns_name += contact.object_name;
    }
    
    visualization_msgs::Marker mk;
    mk.header.stamp = planning_monitor_->lastJointStateUpdate();
    mk.header.frame_id = planning_monitor_->getFrameId();
    mk.ns = ns_name;
    mk.id = count++;
    mk.type = visualization_msgs::Marker::SPHERE;
    mk.action = visualization_msgs::Marker::ADD;
    mk.pose.position.x = contact.pos.x();
    mk.pose.position.y = contact.pos.y();
    mk.pose.position.z = contact.pos.z();
    mk.pose.orientation.w = 1.0;
    
    mk.scale.x = mk.scale.y = mk.scale.z = 0.01;
    
    mk.color.a = 0.6;
    mk.color.r = 1.0;
    mk.color.g = 0.8;
    mk.color.b = 0.04;
    
    //mk.lifetime = ros::Duration(30.0);
    
    vis_marker_publisher_.publish(mk);
  }
	     
  bool getEnvironmentSafety(planning_environment_msgs::GetEnvironmentSafety::Request &req, 
                            planning_environment_msgs::GetEnvironmentSafety::Response &res)
  {
    planning_monitor_->isEnvironmentSafe(res.error_code);
    return true;
  }

  bool getRobotState(planning_environment_msgs::GetRobotState::Request &req, 
                     planning_environment_msgs::GetRobotState::Response &res)
  {
    getCurrentJointState(res.robot_state.joint_state);
    //TODO - does anything else need to go in here?
    //res.robot_state.pose_state.header.stamp = ros::Time::now();
    return true;    
  }

  void getCurrentJointState(sensor_msgs::JointState &joint_state)
  {
    planning_models::KinematicState st(*planning_monitor_->getRobotState());
    // fill in start state with current one
    std::vector<const planning_models::KinematicModel::Joint*> joints;
    planning_monitor_->getKinematicModel()->getJoints(joints);
	    
    joint_state.name.resize(joints.size());
    joint_state.position.resize(joints.size());
    joint_state.header.frame_id = planning_monitor_->getFrameId();
    joint_state.header.stamp = planning_monitor_->lastJointStateUpdate();

    for (unsigned int i = 0 ; i < joints.size() ; ++i)
    {
      joint_state.name[i] = joints[i]->name;
      std::vector<double> tmp;
      st.copyParamsJoint(tmp, joints[i]->name);
      if(!tmp.empty())
        joint_state.position[i] = tmp[0];
    }	    
  }

  bool getJointsInGroup(planning_environment_msgs::GetJointsInGroup::Request &req, 
                        planning_environment_msgs::GetJointsInGroup::Response &res)
  {
    ROS_DEBUG("Looking for joints in group %s",req.group_name.c_str());
    planning_models::KinematicModel::JointGroup *jg = planning_monitor_->getKinematicModel()->getGroup(req.group_name);
    if(jg)
    {
      res.joint_names = jg->jointNames;
      res.error_code.val = res.error_code.SUCCESS;
      return true;
    }
    res.error_code.val = res.error_code.INVALID_GROUP_NAME;
    return true;
  }

  bool getTrajectoryValidity(planning_environment_msgs::GetJointTrajectoryValidity::Request &req, 
                             planning_environment_msgs::GetJointTrajectoryValidity::Response &res) 
  {
    motion_planning_msgs::OrderedCollisionOperations operations;
    std::vector<std::string> child_links;
    planning_monitor_->getChildLinks(req.robot_state.joint_state.name, child_links);
    planning_monitor_->getOrderedCollisionOperationsForOnlyCollideLinks(child_links,req.ordered_collision_operations,operations);
    planning_monitor_->applyOrderedCollisionOperationsToCollisionSpace(operations);
    planning_monitor_->setAllowedContacts(req.allowed_contacts);
    transformAndSetConstraints(req.path_constraints,
                               req.goal_constraints,
                               res.error_code);
    if(res.error_code.val != res.error_code.SUCCESS)
    {
      ROS_ERROR("Could not set constraints");
      return true;
    }
    int flag = getCheckFlag(req);
    if(planning_monitor_->isTrajectoryValid(req.trajectory,
                                            req.robot_state,
                                            (unsigned int) 0, 
                                            (unsigned int) req.trajectory.points.size()-1, 
                                            flag,
                                            false, 
                                            res.error_code,
                                            res.trajectory_error_codes))
    {
      res.error_code.val = res.error_code.SUCCESS;
      ROS_DEBUG("Trajectory is valid");
    }
    visualizeAllowedContactRegions(req.allowed_contacts);
    planning_monitor_->getKinematicModel()->lock();
    planning_monitor_->getEnvironmentModel()->lock();
    planning_monitor_->revertAllowedCollisionToDefault();
    planning_monitor_->clearAllowedContacts();
    planning_monitor_->clearConstraints();
    planning_monitor_->getKinematicModel()->unlock();
    planning_monitor_->getEnvironmentModel()->unlock();
    return true;
  }

  bool getStateValidity(planning_environment_msgs::GetStateValidity::Request &req, 
                        planning_environment_msgs::GetStateValidity::Response &res) 
  {
    motion_planning_msgs::OrderedCollisionOperations operations;
    std::vector<std::string> child_links;
    planning_monitor_->getChildLinks(req.robot_state.joint_state.name, child_links);
    planning_monitor_->getOrderedCollisionOperationsForOnlyCollideLinks(child_links,req.ordered_collision_operations,operations);
    planning_monitor_->applyOrderedCollisionOperationsToCollisionSpace(operations);
    planning_monitor_->setAllowedContacts(req.allowed_contacts);
    transformAndSetConstraints(req.path_constraints,
                               req.goal_constraints,
                               res.error_code);
    if(res.error_code.val != res.error_code.SUCCESS)
    {
      ROS_ERROR("Could not set path constraints");
      return true;
    }
    int flag = getCheckFlag(req);
    if(planning_monitor_->isStateValid(req.robot_state,flag,true,res.error_code))
    {
      res.error_code.val = res.error_code.SUCCESS;
    }

    visualizeAllowedContactRegions(req.allowed_contacts);
    planning_monitor_->getKinematicModel()->lock();
    planning_monitor_->getEnvironmentModel()->lock();
    planning_monitor_->revertAllowedCollisionToDefault();
    planning_monitor_->clearAllowedContacts();
    planning_monitor_->clearConstraints();
    planning_monitor_->getKinematicModel()->unlock();
    planning_monitor_->getEnvironmentModel()->unlock();
    return true;
  }

  bool transformAndSetConstraints(motion_planning_msgs::Constraints &path_constraints, 
                                  motion_planning_msgs::Constraints &goal_constraints,
                                  motion_planning_msgs::ArmNavigationErrorCodes &error_code)
  {
    if(!planning_monitor_->transformConstraintsToFrame(goal_constraints, 
                                                       planning_monitor_->getFrameId(), 
                                                       error_code))
      return true;

    if(!planning_monitor_->transformConstraintsToFrame(path_constraints, 
                                                       planning_monitor_->getFrameId(),
                                                       error_code))
      return true;
          
    planning_monitor_->setPathConstraints(path_constraints,error_code);
    planning_monitor_->setGoalConstraints(goal_constraints,error_code);

    error_code.val = error_code.SUCCESS;
    return true;    
  }


  bool setConstraints(planning_environment_msgs::SetConstraints::Request &req, 
                      planning_environment_msgs::SetConstraints::Response &res)
  {
    if(!planning_monitor_->transformConstraintsToFrame(req.goal_constraints, 
                                                       planning_monitor_->getFrameId(), 
                                                       res.error_code))
      return true;

    if(!planning_monitor_->transformConstraintsToFrame(req.path_constraints, 
                                                       planning_monitor_->getFrameId(),
                                                       res.error_code))
      return true;
          
    planning_monitor_->setPathConstraints(req.path_constraints,res.error_code);
    planning_monitor_->setGoalConstraints(req.goal_constraints,res.error_code);

    res.error_code.val = res.error_code.SUCCESS;
    res.path_constraints = req.path_constraints;
    res.goal_constraints = req.goal_constraints;
    return true;
  }

  bool GetExecutionSafety(planning_environment_msgs::GetJointTrajectoryValidity::Request &req, 
                          planning_environment_msgs::GetJointTrajectoryValidity::Response &res)
  {
    motion_planning_msgs::OrderedCollisionOperations operations;
    std::vector<std::string> child_links;
    planning_monitor_->getChildLinks(req.robot_state.joint_state.name,child_links);
    planning_monitor_->getOrderedCollisionOperationsForOnlyCollideLinks(child_links,req.ordered_collision_operations,operations);
    planning_monitor_->applyOrderedCollisionOperationsToCollisionSpace(operations);
    planning_monitor_->setAllowedContacts(req.allowed_contacts);
    transformAndSetConstraints(req.path_constraints,
                               req.goal_constraints,
                               res.error_code);
    if(res.error_code.val != res.error_code.SUCCESS)
    {
      ROS_ERROR("Could not set constraints");
      return true;
    }
    if(!planning_monitor_->isEnvironmentSafe(res.error_code))
      return true;

    int current_position_index(0);        
    // we don't want to check the part of the path that was already executed
    current_position_index = planning_monitor_->closestStateOnTrajectory(req.trajectory,
                                                                         req.robot_state,
                                                                         current_position_index, 
                                                                         req.trajectory.points.size() - 1, 
                                                                         planning_monitor_->getRobotState(),res.error_code);
    if (current_position_index < 0)
    {
      ROS_WARN("Unable to identify current state in path");
      current_position_index = 0;
    }
		    
    bool valid = false;
    planning_monitor_->setAllowedContacts(allowed_contacts_);

    valid = planning_monitor_->isTrajectoryValid(req.trajectory,
                                                 req.robot_state,
                                                 current_position_index, 
                                                 req.trajectory.points.size() - 1, 
                                                 planning_environment::PlanningMonitor::PATH_CONSTRAINTS_TEST | planning_environment::PlanningMonitor::COLLISION_TEST,
                                                 false,res.error_code,res.trajectory_error_codes);

    if(velocity_history_index_ >= velocity_history_.size())
    {		
      double sum = velocity_history_.sum();
      if (sum < 1e-3)
      {
        ROS_INFO("The total velocity of the robot over the last %d samples is %f. Self-preempting...", (int)velocity_history_.size(), sum);
        res.error_code.val = res.error_code.JOINTS_NOT_MOVING;
      }
    }

    planning_monitor_->getKinematicModel()->lock();
    planning_monitor_->getEnvironmentModel()->lock();
    planning_monitor_->revertAllowedCollisionToDefault();
    planning_monitor_->clearAllowedContacts();
    planning_monitor_->getKinematicModel()->unlock();
    planning_monitor_->getEnvironmentModel()->unlock();

    return true;
  }

  int getCheckFlag(const planning_environment_msgs::GetJointTrajectoryValidity::Request &req)
  {
    int result;
    if(req.check_collisions)
      result = result | planning_environment::PlanningMonitor::COLLISION_TEST;
    if(req.check_path_constraints)
      result = result | planning_environment::PlanningMonitor::PATH_CONSTRAINTS_TEST;
    if(req.check_goal_constraints)
      result = result | planning_environment::PlanningMonitor::GOAL_CONSTRAINTS_TEST;
    if(req.check_joint_limits)
      result = result | planning_environment::PlanningMonitor::JOINT_LIMITS_TEST;
    if(req.check_full_trajectory)
      result = result | planning_environment::PlanningMonitor::CHECK_FULL_TRAJECTORY;
    return result;
  }

  int getCheckFlag(const planning_environment_msgs::GetStateValidity::Request &req)
  {
    int result;
    if(req.check_collisions)
      result = result | planning_environment::PlanningMonitor::COLLISION_TEST;
    if(req.check_path_constraints)
      result = result | planning_environment::PlanningMonitor::PATH_CONSTRAINTS_TEST;
    if(req.check_goal_constraints)
      result = result | planning_environment::PlanningMonitor::GOAL_CONSTRAINTS_TEST;
    if(req.check_joint_limits)
      result = result | planning_environment::PlanningMonitor::JOINT_LIMITS_TEST;
    return result;
  }

  void jointStateCallback(const sensor_msgs::JointStateConstPtr &joint_state)
  {
    velocity_history_[velocity_history_index_ % velocity_history_.size()] = planning_monitor_->getTotalVelocity();
    velocity_history_index_++;
  }

  void visualizeAllowedContactRegions(const std::vector<motion_planning_msgs::AllowedContactSpecification> &allowed_contacts)
  {
    static int count = 0;
    visualization_msgs::MarkerArray mk;
    mk.markers.resize(allowed_contacts.size());
    for(unsigned int i=0; i < allowed_contacts.size(); i++) 
    { 
      bool valid_shape = true;
      mk.markers[i].header.stamp = ros::Time::now();
      mk.markers[i].header.frame_id = allowed_contacts[i].pose_stamped.header.frame_id;
      mk.markers[i].ns = allowed_contacts[i].name;
      mk.markers[i].id = count++;
      if(allowed_contacts[i].shape.type == geometric_shapes_msgs::Shape::SPHERE)
      {        
        mk.markers[i].type = visualization_msgs::Marker::SPHERE;
        if(allowed_contacts[i].shape.dimensions.size() >= 1)
          mk.markers[i].scale.x = mk.markers[i].scale.y = mk.markers[i].scale.z = allowed_contacts[i].shape.dimensions[0];
        else
          valid_shape = false;
      }      
      else if (allowed_contacts[i].shape.type == geometric_shapes_msgs::Shape::BOX)
      {
        mk.markers[i].type = visualization_msgs::Marker::CUBE;
        if(allowed_contacts[i].shape.dimensions.size() >= 3)
        {
          mk.markers[i].scale.x = allowed_contacts[i].shape.dimensions[0];
          mk.markers[i].scale.y = allowed_contacts[i].shape.dimensions[1];
          mk.markers[i].scale.z = allowed_contacts[i].shape.dimensions[2];
        }
        else
          valid_shape = false;
      }
      else if (allowed_contacts[i].shape.type == geometric_shapes_msgs::Shape::CYLINDER)
      {
        mk.markers[i].type = visualization_msgs::Marker::CYLINDER;
        if(allowed_contacts[i].shape.dimensions.size() >= 2)
        {
          mk.markers[i].scale.x = allowed_contacts[i].shape.dimensions[0];
          mk.markers[i].scale.y = allowed_contacts[i].shape.dimensions[0];
          mk.markers[i].scale.z = allowed_contacts[i].shape.dimensions[1];
        }
        else
          valid_shape = false;
      }
      else
      {
        mk.markers[i].scale.x = mk.markers[i].scale.y = mk.markers[i].scale.z = 0.01;
        valid_shape = false;
      }        

      mk.markers[i].action = visualization_msgs::Marker::ADD;
      mk.markers[i].pose = allowed_contacts[i].pose_stamped.pose;  
      if(!valid_shape)
      {
        mk.markers[i].scale.x = mk.markers[i].scale.y = mk.markers[i].scale.z = 0.01;
        mk.markers[i].color.a = 0.5;
        mk.markers[i].color.r = 1.0;
        mk.markers[i].color.g = 0.04;
        mk.markers[i].color.b = 0.04;
      }
      else
      {
        mk.markers[i].color.a = 0.5;
        mk.markers[i].color.r = 0.04;
        mk.markers[i].color.g = 1.0;
        mk.markers[i].color.b = 0.04;
      }  
      //mk.markers[i].lifetime = ros::Duration(30.0);  
    }
    allowed_contact_regions_publisher_.publish(mk);
  }

private:
	
  ros::NodeHandle root_handle_, private_handle_;
  EnvironmentServerSetup &setup_;
  planning_environment::PlanningMonitor *planning_monitor_;
  tf::TransformListener *tf_;	

  bool show_collisions_;
  bool allow_valid_collisions_;	
  std::vector<collision_space::EnvironmentModel::AllowedContact> allowed_contacts_;

  ros::Publisher vis_marker_publisher_;
  ros::ServiceServer get_trajectory_validity_service_;
  ros::ServiceServer get_env_safety_service_;
  ros::ServiceServer get_execution_safety_service_;
  ros::ServiceServer set_constraints_service_;
  ros::ServiceServer get_joints_in_group_service_;
  ros::ServiceServer get_robot_state_service_;
  ros::ServiceServer get_state_validity_service_;
  ros::Publisher allowed_contact_regions_publisher_;

  ros::Subscriber joint_state_subscriber_;

  std::valarray<double> velocity_history_;
  unsigned int velocity_history_index_;

};    
}

int main(int argc, char** argv)
{
  ros::init(argc, argv, "environment_monitor");
  planning_environment::EnvironmentServerSetup setup;
  if (setup.configure())
  {
    planning_environment::EnvironmentServer environment_monitor(setup);
    ros::spin();
  }    
  return 0;
}
